# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
satmad
=========
Satellite Mission Analysis and Design in Python.
"""
__version__ = "0.1"
